@@ -191,7 +191,7 @@ class ClusteringOptimizer:
 		return [
 			{
 				'k': k, 'distortion': self._distortions[k], 'inertia': self._inertias[k],
-				'silhouette_score': self._silhouette_scores
+				'silhouette_score': self._silhouette_scores[k]
 			}
 			for k in range(self._min_k, self._max_k + 1)
 		]
