@@ -7,7 +7,6 @@ from sklearn.metrics import silhouette_score
 from pyspark.sql import functions as f
 from pyspark.sql.types import DoubleType, StringType
 from pyspark.sql import DataFrame as SparkDF, Column as SparkColumn
-from pyspark.sql.window import Window
 from pyspark.ml.clustering import KMeans as SparkKMeans
 from pyspark.ml.feature import VectorAssembler
 from pyspark.ml.evaluation import ClusteringEvaluator
@@ -223,62 +222,92 @@ class KMeans:
 		"""
 		return self._spark_model.clusterCenters()
 
-	def _transform_spark(self, X, prefix='distance_', keep_columns=False):
+	@staticmethod
+	def _get_keep_columns(X, keep_columns):
+		if isinstance(keep_columns, str):
+			keep_columns = [keep_columns]
+		elif isinstance(keep_columns, list):
+			pass
+		elif keep_columns == True:
+			keep_columns = list(X.columns)
+		elif keep_columns == False or keep_columns is None:
+			keep_columns = []
+		else:
+			raise TypeError(f'keep_columns of type {type(keep_columns)} is not supported!')
+		return keep_columns
+
+	def _transform_and_predict_spark(self, X, add_distances, keep_prediction, distance_prefix, cluster_prefix, keep_columns):
+		assembled = self._spark_assembler.transform(X)
+		transformed = self._spark_model.transform(assembled)
+
+		keep_columns = self._get_keep_columns(X=X, keep_columns=keep_columns)
+		transformed = transformed.select(*[
+			col for col in transformed.columns
+			if (col in transformed.columns and col not in X.columns) or col in keep_columns
+		])
+
+		if add_distances:
+			centers = self.cluster_centers
+
+			@f.udf(DoubleType())
+			def _get_distance(features, cluster):
+				"""
+				:type features: np_array
+				:rtype: SparkColumn
+				"""
+				center = centers[cluster]
+				return float(Vectors.squared_distance(features, center) ** 0.5)
+
+			transformed = (
+				transformed
+				.select(*transformed.columns, *[
+					_get_distance('features', f.lit(cluster)).alias(f'{distance_prefix}{cluster + 1}')
+					for cluster in range(len(centers))
+				])
+			)
+
+		transformed = transformed.drop('features')
+
+		if not keep_prediction:
+			transformed = transformed.drop('prediction')
+		else:
+			@f.udf(StringType())
+			def _convert_prediction_to_string(x):
+				return f'{cluster_prefix}{x + 1}'
+
+			transformed = transformed.withColumn('prediction', _convert_prediction_to_string('prediction'))
+
+		return transformed
+
+	def _predict_spark(self, X, prefix='cluster_', keep_columns=False, add_distance=False, distance_prefix='distance_'):
 		"""
 		:type X: SparkDF
 		:type prefix: str
+		:type keep_columns: bool or NoneType or list[str] or str
+		:type add_distance: bool
+		:type distance_prefix: str
 		:rtype: SparkDF
 		"""
-		assembled = self._spark_assembler.transform(X)
-		centers = self.cluster_centers
+		return self._transform_and_predict_spark(
+			X=X, cluster_prefix=prefix, keep_columns=keep_columns, keep_prediction=True,
+			add_distances=add_distance, distance_prefix=distance_prefix
+		)
 
-		@f.udf(DoubleType())
-		def _get_distance(features, cluster):
-			"""
-			:type features: np_array
-			:rtype: SparkColumn
-			"""
-			center = centers[cluster]
-			return float(Vectors.squared_distance(features, center) ** 0.5)
-
-		if isinstance(keep_columns, (list, str)):
-			if isinstance(keep_columns, str):
-				keep_columns = [keep_columns]
-		elif keep_columns:
-			keep_columns = [col for col in X.columns if col != 'features']
-		else:
-			keep_columns = []
-
-		transformed = assembled.select(*keep_columns, *[
-			_get_distance('features', f.lit(cluster)).alias(f'{prefix}{cluster + 1}')
-			for cluster in range(len(centers))
-		])
-		return transformed
-
-	def _predict_spark(self, X, prefix='cluster_', keep_columns=False):
+	def _transform_spark(
+			self, X, prefix='distance_', keep_columns=False, add_prediction=False, cluster_prefix='cluster_'
+	):
 		"""
 		:type X: SparkDF
+		:type prefix: str
+		:type keep_columns: bool or NoneType or list[str] or str
+		:type add_prediction: bool
+		:type cluster_prefix: str
 		:rtype: SparkDF
 		"""
-		assembled = self._spark_assembler.transform(X)
-		result = self._spark_model.transform(assembled)
-
-		@f.udf(StringType())
-		def _convert_prediction_to_string(x):
-			return f'{prefix}{x + 1}'
-		result = result.withColumn('prediction', _convert_prediction_to_string('prediction'))
-
-		if keep_columns:
-			if isinstance(keep_columns, str):
-				keep_columns = [keep_columns]
-
-			if isinstance(keep_columns, list):
-				columns = keep_columns + ['prediction']
-				return result.select(*columns)
-			else:
-				return result.drop('features')
-		else:
-			return result.select('prediction')
+		return self._transform_and_predict_spark(
+			X=X, distance_prefix=prefix, keep_columns=keep_columns, add_distances=True,
+			keep_prediction=add_prediction, cluster_prefix=cluster_prefix
+		)
 
 	# GENERAL METHODS #
 
@@ -318,7 +347,7 @@ class KMeans:
 	def transform(self, X, keep_columns=False):
 		"""
 		:type X: SparkDF or PandasDF
-		:type keep_columns: bool
+		:type keep_columns: bool or list[str] or str
 		"""
 		if self._env == 'spark':
 			return self._transform_spark(X=X, keep_columns=keep_columns)
@@ -366,22 +395,18 @@ class KMeans:
 		:type keep_columns: bool or str or list[str]
 		:rtype: PandasDF
 		"""
-		transformed = self.transform(X=X, keep_columns=keep_columns)
-		predictions = self.predict(X=X, return_dataframe=False, keep_columns=False)
 
-		if self._env == 'python':
-			transformed['prediction'] = predictions
-
-		elif self._env == 'spark':
-			w = Window.orderBy(f.monotonically_increasing_id())
-			transformed = transformed.withColumn('_index', f.row_number().over(w))
-			predictions = predictions.withColumn('_index', f.row_number().over(w))
-			transformed = transformed.join(predictions, on='_index', how='inner').drop('_index')
+		if isinstance(X, SparkDF):
+			return self._transform_and_predict_spark(
+				X=X, add_distances=True, keep_prediction=True, distance_prefix='distance_', cluster_prefix='cluster_',
+				keep_columns=keep_columns
+			)
 
 		else:
-			raise RuntimeError(f'unknown environment {self._env}')
-
-		return transformed
+			transformed = self._transform_python(X=X, keep_columns=True, prefix='distance_')
+			keep_columns = self._get_keep_columns(X=X, keep_columns=keep_columns)
+			keep_columns = keep_columns + [col for col in transformed.columns if col not in X.columns]
+			return self._predict_python(X=transformed, return_dataframe=True, keep_columns=keep_columns)
 
 	@property
 	def n_clusters(self):
