@@ -5,7 +5,7 @@ from .get_month import get_month_name, get_month_num, get_months_list
 from .get_number_of_days import get_number_of_days
 from .get_quarter import get_quarter
 from .get_today import get_today, get_today_str, get_date
-from .parse_date import parse_date
+from .parse_date import parse_date, find_date_format, find_date_and_format
 from .YearMonth import YearMonth
 from .YearQuarter import YearQuarter
 from .get_first_day_of_month import get_first_day_of_month, get_last_day_of_month
